import numpy as np
import matplotlib.pyplot as plt
from qutip import * 

import scipy.sparse as spr
import block_diagonalize
import itertools
import multiprocessing as mp
import time
import os
import pickle


def run_fun(params):
    G = params[0]
    Δ = params[1]
    Utilde = params[2]
    U = Utilde/N
    
    H = -Δ*a.dag()*a + G/2 *(a.dag()**2 +a**2) + U/2 *a.dag()**2*a**2
    LL = liouvillian(H, c_ops)

    P, block_bfs, block_sizes = block_diagonalize.PermMat(LL)
    bd_L = np.dot(P, np.dot(LL.data, np.transpose(P)))
    num_blocks, blocks_list, bl_indices = block_diagonalize.get_blocks(bd_L)   

    done=False
    for i in range(int(len(blocks_list))):
        block = blocks_list[i]
        evals, evecs = Qobj(block).eigenstates()    

        ss_block_form = evecs[-1]
        evec2 = spr.dok_matrix((LL.shape[0], 1), dtype=np.complex)
        evec2[bl_indices[i]:bl_indices[i]+blocks_list[i].shape[0]] = ss_block_form.data
        evec2 = evec2.tocsr()
        evec2 = np.dot(np.transpose(P),evec2)
        ss = Qobj(evec2, dims=[LL.dims[0], [1]])
        ss = vector_to_operator(ss)

        if np.real(ss.tr())>0.05:
            ss=ss+ss.dag()
            ss/=(ss.tr())       
            if expect(parity, ss)>0.5:
                num_even = expect(ss, a.dag()*a)
                gap_even = evals[-2]
            else:
                num_odd = expect(ss, a.dag()*a)                
                gap_odd = evals[-2]
        elif done==False:
            gap_even_odd = np.real(evals[-1])+1.j*np.abs(np.imag(evals[-1]))
            gap_odd_even = np.real(evals[-1])-1.j*np.abs(np.imag(evals[-1]))       
            done=True
                
    return [num_even, num_odd, gap_even, gap_odd, gap_even_odd, gap_odd_even]


if __name__=="__main__":
    a=destroy(50)

    parity=1.j*np.pi*a.dag()*a
    parity=parity.expm()

    ηtilde = 1
    Utilderange = np.linspace(0,20,81)
    Δrange = np.linspace(0,20,80)
    Grange = np.linspace(0,14,20, retstep=True)
    paramlist = list(itertools.product(Grange, Δrange, Utilderange))

    Δ = 10
    N = 15
    η = ηtilde/N
    c_ops=[np.sqrt(η)*a**2]

    #perform parallel evaluation
    t1 = time.time()
    pool = mp.Pool()
    results = pool.map(run_fun,paramlist)
    print(time.time()-t1)

    with open("DeltaU_multiG_scan.pckl", "wb") as fp:
        pickle.dump([Grange, Δrange, Utilderange, results], fp)
        
        
        
        
        